# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import os, shutil
import tempfile
import unittest
import sys
import time
import json
import random
import logging
from datetime import datetime, timedelta
import six
from six.moves import queue
from slapos.grid.promise import (interface, PromiseLauncher, PromiseProcess,
                                 PromiseError, PROMISE_CACHE_FOLDER_NAME)
from slapos.grid.promise.generic import (GenericPromise, TestResult, AnomalyResult,
                                         PromiseQueueResult, PROMISE_STATE_FOLDER_NAME,
                                         PROMISE_HISTORY_RESULT_FOLDER_NAME,
                                         PROMISE_RESULT_FOLDER_NAME,
                                         PROMISE_PARAMETER_NAME)

class TestSlapOSPromiseMixin(unittest.TestCase):

  def setUp(self):
    self.partition_dir = tempfile.mkdtemp()
    self.plugin_dir = os.path.join(self.partition_dir, 'plugins')
    self.legacy_promise_dir = os.path.join(self.partition_dir, 'promise')
    self.log_dir = os.path.join(self.partition_dir, 'log')
    os.mkdir(self.plugin_dir)
    os.mkdir(self.log_dir)
    os.makedirs('%s/.slapgrid/promise' % self.partition_dir)
    os.mkdir(self.legacy_promise_dir)
    self.partition_id = "slappart0"
    self.computer_id = "COMP-1234"

  def writeInit(self):
    with open(os.path.join(self.plugin_dir, '__init__'), 'w') as f:
      f.write('')
    os.chmod(os.path.join(self.plugin_dir, '__init__'), 0o644)
    if sys.path[0] != self.plugin_dir:
      sys.path[0:0] = [self.plugin_dir]

  def tearDown(self):
    if os.path.exists(self.partition_dir):
      shutil.rmtree(self.partition_dir)
    if sys.path[0] == self.plugin_dir:
      del sys.path[0]

  def configureLauncher(self, save_method=None, timeout=5, master_url="", debug=False,
      run_list=[], uid=None, gid=None, enable_anomaly=False, force=False,
      logdir=True, dry_run=False):
    parameter_dict = {
      'promise-timeout': timeout,
      'promise-folder': self.plugin_dir,
      'legacy-promise-folder': self.legacy_promise_dir,
      'log-folder': self.log_dir if logdir else None,
      'partition-folder': self.partition_dir,
      'master-url': master_url,
      'partition-cert': "",
      'partition-key': "",
      'partition-id': self.partition_id,
      'computer-id': self.computer_id,
      'debug': debug,
      'check-anomaly': enable_anomaly,
      'force': force,
      'run-only-promise-list': run_list,
      'uid': uid,
      'gid': gid
    }

    self.launcher = PromiseLauncher(
      config=parameter_dict,
      logger=logging.getLogger('slapos.test.promise'),
      dry_run=dry_run
    )
    if save_method:
      self.launcher._savePromiseResult = save_method

  def genPromiseConfigDict(self, promise_name):
    return {
      'log-folder': None,
      'partition-folder': self.partition_dir,
      'master-url': "https://master.url.com",
      'partition-cert': "",
      'partition-key': "",
      'partition-id': self.partition_id,
      'computer-id': self.computer_id,
      'debug': False,
      'name': promise_name,
      'path': os.path.join(self.plugin_dir, promise_name),
      'queue': queue.Queue(),
    }

  def createPromiseProcess(self, promise_name, check_anomaly=False, wrap=False):

    promise_path = os.path.join(self.plugin_dir, promise_name)
    return PromiseProcess(
      self.partition_dir,
      promise_name,
      promise_path,
      logger=logging.getLogger('slapos.test.promise'),
      argument_dict=self.genPromiseConfigDict(promise_name),
      check_anomaly=check_anomaly,
      wrap=wrap,
    )

  def writeFile(self, path, content, mode=0o644):
    with open(path, 'w') as f:
      f.write(content)
    os.chmod(path, mode)

  def generatePromiseScript(self, name, success=True, failure_count=1, content="",
    periodicity=0.03, is_tested=True, with_anomaly=True):
    promise_content = """from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise import GenericPromise

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):
    GenericPromise.__init__(self, config)
    self.setPeriodicity(minute=%(periodicity)s)
    if not %(is_tested)s:
      self.setTestLess()
    if not %(with_anomaly)s:
      self.setAnomalyLess()

  def sense(self):
    %(content)s

    if not %(success)s:
      self.logger.error("failed")
    else:
      self.logger.info("success")

  def anomaly(self):
    return self._anomaly(latest_minute=%(periodicity)s, failure_amount=%(failure_amount)s)

  def test(self):
    return self._test(latest_minute=%(periodicity)s, failure_amount=%(failure_amount)s)

""" % {'success': success, 'content': content, 'failure_amount': failure_count,
       'periodicity': periodicity, 'is_tested': is_tested, 'with_anomaly': with_anomaly}

    with open(os.path.join(self.plugin_dir, name), 'w') as f:
      f.write(promise_content)

  def assertSuccessResult(self, name):
    expected_result = """{
  "result": {
    "failed": false,
    "message": "success",
    "type": "Test Result"
  },
  "path": "%(promise_dir)s/%(name)s.py",
  "name": "%(name)s.py",
  "title": "%(name)s"
}"""

    # first promise
    state_file = os.path.join(self.partition_dir, PROMISE_RESULT_FOLDER_NAME, '%s.status.json' % name)
    self.assertTrue(os.path.exists(state_file))
    with open(state_file) as f:
      result_dict = json.loads(f.read())
      result_dict['result'].pop('date')
      # execution time of a promise can vary from 0.05 to 0.2 in a busy test machine
      # it makes no sense to test it
      execution_time = result_dict.pop('execution-time')
      self.assertTrue(execution_time > 0.0 and execution_time <=1.0)
      expected_dict = expected_result % {'promise_dir': self.plugin_dir, 'name': name}
      self.assertEqual(json.loads(expected_dict), result_dict)


  def assertSuccessHistoryResult(self, name, expected_history=None):
    if not expected_history:
      expected_history = """{
	"data": [{
		"failed": false,
		"message": "success",
		"name": "%(name)s.py",
		"status": "OK",
		"title": "%(name)s"
	}]
}"""

    history_file = os.path.join(self.partition_dir, PROMISE_HISTORY_RESULT_FOLDER_NAME, '%s.history.json' % name)
    self.assertTrue(os.path.exists(history_file))
    with open(history_file) as f:
      result_dict = json.loads(f.read())
      result_dict.pop('date')
      for entry in result_dict["data"]:
        d = entry.pop("date")
        self.assertEqual(d, entry.pop("change-date"))
        # execution time of a promise can vary from 0.05 to 0.2 in a busy test machine
        # it makes no sense to test it
        execution_time = entry.pop("execution-time")
        self.assertTrue(execution_time > 0.0 and execution_time <=1.0)

      expected_dict = expected_history % {'name': name}
      self.assertEqual(json.loads(expected_dict), result_dict)

  def assertSuccessStatsResult(self, success=1, error=0, expected_stats=None):
    if not expected_stats:
      expected_stats = """{
	"data": ["Date, Success, Error, Warning",
                "__DATE__, %(success)s, %(error)s, "
	]
}"""

    stats_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME, 'promise_stats.json')
    self.assertTrue(os.path.exists(stats_file))
    with open(stats_file) as f:
      result_dict = json.loads(f.read())
      result_dict.pop('date')
      expected_dict = expected_stats % {'success': success, "error": error}
      copy = result_dict["data"]
      for nline in range(1, len(result_dict["data"])):
        line = result_dict["data"][nline]
        result_dict["data"][nline] = "__DATE__,%s" % ",".join(line.split(',')[1:])
      self.assertEqual(json.loads(expected_dict), result_dict)

class TestSlapOSPromiseLauncher(TestSlapOSPromiseMixin):

  def test_promise_match_interface(self):
    promise_name = 'my_promise.py'
    self.writeInit()
    self.generatePromiseScript(promise_name)
    promise_process = self.createPromiseProcess(promise_name)

    promise_module = promise_process._loadPromiseModule()

  def test_promise_match_interface_bad_name(self):
    promise_name = 'my_promise_no_py'
    self.writeInit()
    self.generatePromiseScript(promise_name)
    promise_process = self.createPromiseProcess(promise_name)

    with self.assertRaises(ImportError) as exc:
      promise_module = promise_process._loadPromiseModule()
    self.assertEqual(str(exc.exception), 'No module named %s' %
                                          ("'%s'" % promise_name if six.PY3 else
                                           promise_name))

  def test_promise_match_interface_no_implement(self):
    promise_name = 'my_promise_noimplement.py'
    promise_content = """from slapos.grid.promise import GenericPromise

class RunPromise(GenericPromise):

  def __init__(self, config):
    GenericPromise.__init__(self, config)

  def sense(self):
    pass
    
"""
    promise_path = os.path.join(self.plugin_dir, promise_name)
    self.writeFile(promise_path, promise_content)
    self.writeInit()
    promise_process = self.createPromiseProcess(promise_name)

    with self.assertRaises(RuntimeError) as exc:
      promise_module = promise_process._loadPromiseModule()
    message = "RunPromise class in my_promise_noimplement.py must implement" \
      " 'IPromise' interface. @implementer(interface.IPromise) is missing ?"
    self.assertEqual(str(exc.exception), message)

  def test_promise_match_interface_no_generic(self):
    promise_name = 'my_promise_nogeneric.py'
    promise_content = """from zope.interface import implementer
from slapos.grid.promise import interface

@implementer(interface.IPromise)
class RunPromise(object):

  def __init__(self, config):
    pass

  def sense(self):
    pass
    
"""
    promise_path = os.path.join(self.plugin_dir, promise_name)
    self.writeFile(promise_path, promise_content)
    self.writeInit()
    promise_process = self.createPromiseProcess(promise_name)

    with self.assertRaises(RuntimeError) as exc:
      promise_module = promise_process._loadPromiseModule()

    self.assertEqual(str(exc.exception), 'RunPromise class is not a subclass of GenericPromise class.')

  def test_promise_match_interface_no_sense(self):
    promise_name = 'my_promise_nosense.py'
    promise_content = """from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise import GenericPromise

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def __init__(self, config):
    pass

  def noSenseMethod(self):
    pass
    
"""
    promise_path = os.path.join(self.plugin_dir, promise_name)
    self.writeFile(promise_path, promise_content)
    self.writeInit()
    promise_process = self.createPromiseProcess(promise_name)

    with self.assertRaises(TypeError) as exc:
      promise_module = promise_process._loadPromiseModule()
      promise = promise_module.RunPromise({})

    expected_assertion_message = "Can't instantiate abstract class RunPromise with abstract method sense"
    if sys.version_info < (3, 9):
      expected_assertion_message = "Can't instantiate abstract class RunPromise with abstract methods sense"

    self.assertEqual(str(exc.exception), expected_assertion_message)

  def test_promise_extra_config(self):
    promise_name = 'my_promise_extra.py'
    config_dict = {'foo': 'bar', 'my-config': 4522111,
                   'text': 'developers \ninformation, sample\n\ner'}
    promise_content = """from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise import GenericPromise

%(config_name)s = %(config_content)s

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def sense(self):
    pass
    
""" % {'config_name': PROMISE_PARAMETER_NAME,
       'config_content': config_dict}
    promise_path = os.path.join(self.plugin_dir, promise_name)
    self.writeFile(promise_path, promise_content)
    self.writeInit()
    promise_process = self.createPromiseProcess(promise_name)
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(promise_process.argument_dict)

    self.assertEqual(promise.getConfig('foo'), 'bar')
    self.assertEqual(promise.getConfig('my-config'), 4522111)
    self.assertEqual(promise.getConfig('text'), config_dict['text'])

  def test_promise_extra_config_reserved_name(self):
    promise_name = 'my_promise_extra.py'
    from collections import OrderedDict
    config_dict = OrderedDict([('name', 'bar'), ('my-config', 4522111)])
    promise_content = """from collections import OrderedDict

from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise import GenericPromise

%(config_name)s = %(config_content)s

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def sense(self):
    pass
    
""" % {'config_name': PROMISE_PARAMETER_NAME,
       'config_content': config_dict}
    promise_path = os.path.join(self.plugin_dir, promise_name)
    self.writeFile(promise_path, promise_content)
    self.writeInit()
    promise_process = self.createPromiseProcess(promise_name)

    with self.assertRaises(ValueError) as exc:
      promise_module = promise_process._loadPromiseModule()
    self.assertEqual(str(exc.exception), "Extra parameter name 'name' cannot be used.\n%s" % config_dict)

  def test_runpromise(self):
    promise_name = 'my_promise.py'
    self.configureLauncher()
    self.generatePromiseScript(promise_name, success=True)
    state_folder = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    self.assertTrue(os.path.exists(state_folder))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))

    self.assertSuccessResult("my_promise")
    self.assertSuccessHistoryResult("my_promise")
    self.assertSuccessStatsResult(1)

  def test_runpromise_multiple(self):
    promise_name = 'my_promise.py'
    second_name = 'my_second_promise.py'
    self.configureLauncher()
    self.generatePromiseScript(promise_name, success=True)
    self.generatePromiseScript(second_name, success=True)
    state_folder = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    self.assertTrue(os.path.exists(state_folder))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))

    self.assertSuccessResult("my_promise")
    self.assertSuccessResult("my_second_promise")

    self.assertSuccessHistoryResult("my_promise")
    self.assertSuccessHistoryResult("my_second_promise")
    self.assertSuccessStatsResult(2)

  def test_runpromise_multiple_times_same_promise(self):
    promise_name = 'my_promise.py'
    self.configureLauncher()
    self.generatePromiseScript(promise_name, success=True)
    state_folder = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    time.sleep(1)
    self.launcher.run()
    time.sleep(1)
    self.launcher.run()
    time.sleep(1)
    self.assertTrue(os.path.exists(state_folder))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))

    self.assertSuccessResult("my_promise")
    self.assertSuccessHistoryResult("my_promise", expected_history = """{
        "data": [{
                "failed": false,
                "message": "success",
                "name": "%(name)s.py",
                "status": "OK",
                "title": "%(name)s"
        },{
                "failed": false,
                "message": "success",
                "status": "OK"
        }]
}""")

    self.assertSuccessStatsResult(1, expected_stats = """{
        "data": ["Date, Success, Error, Warning",
                "__DATE__, %(success)s, %(error)s, ",
                "__DATE__, %(success)s, %(error)s, ",
                "__DATE__, %(success)s, %(error)s, "
        ]
}""")

  def test_runpromise_multiple_times_same_promise_with_failure(self):
    promise_name = 'my_promise.py'
    self.configureLauncher()
    self.generatePromiseScript(promise_name, success=True)
    state_folder = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    time.sleep(1)
    self.generatePromiseScript(promise_name, success=False)
    time.sleep(1)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    time.sleep(1)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    time.sleep(1)
    self.assertTrue(os.path.exists(state_folder))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))

    self.assertSuccessHistoryResult("my_promise", expected_history = """{
        "data": [{
                "failed": false,
                "message": "success",
                "name": "%(name)s.py",
                "status": "OK",
                "title": "%(name)s"
        },{
                "failed": true,
                "message": "failed",
                "status": "ERROR"
        }]
}""")
    self.assertSuccessStatsResult(1, expected_stats = """{
        "data": ["Date, Success, Error, Warning",
                "__DATE__, %(success)s, %(error)s, ",
                "__DATE__, 0, 1, ",
                "__DATE__, 0, 1, "
        ]
}""")

  def test_runpromise_multiple_times_same_promise_with_flaky_failures(self):
    promise_name = 'my_promise.py'
    self.configureLauncher()
    self.generatePromiseScript(promise_name, success=True)
    state_folder = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    time.sleep(1)
    self.generatePromiseScript(promise_name, success=False)
    time.sleep(1)
    with self.assertRaises(PromiseError):
      self.launcher.run()
    time.sleep(1)
    self.generatePromiseScript(promise_name, success=True)
    time.sleep(1)
    self.launcher.run()
    self.assertTrue(os.path.exists(state_folder))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))
    self.maxDiff = None
    self.assertSuccessHistoryResult("my_promise", expected_history = """{
        "data": [{
                "failed": false,
                "message": "success",
                "name": "%(name)s.py",
                "status": "OK",
                "title": "%(name)s"
        },{
                "failed": true,
                "message": "failed",
                "status": "ERROR"
        },{
                "failed": false,
                "message": "success",
                "status": "OK"
        }]
}""")
    self.assertSuccessStatsResult(1, expected_stats = """{
        "data": ["Date, Success, Error, Warning",
                "__DATE__, %(success)s, %(error)s, ",
                "__DATE__, 0, 1, ",
                "__DATE__, %(success)s, %(error)s, "
        ]
}""")



  def test_runpromise_no_logdir(self):
    promise_name = 'my_promise.py'
    # no promise log output dir
    self.configureLauncher(logdir=False)
    self.generatePromiseScript(promise_name, success=True)
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    self.assertTrue(os.path.exists(state_file))
    self.assertFalse(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))
    self.assertSuccessHistoryResult("my_promise")
    self.assertSuccessStatsResult(1)

  def test_runpromise_savemethod(self):
    promise_name = 'my_promise.py'
    def test_method(result):
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, TestResult))
      self.assertTrue(result.execution_time != 0)
      self.assertEqual(result.title, 'my_promise')
      self.assertEqual(result.name, promise_name)
      self.assertEqual(result.path, os.path.join(self.plugin_dir, promise_name))
      self.assertEqual(result.item.message, "success")
      self.assertEqual(result.item.hasFailed(), False)
      self.assertTrue(isinstance(result.item.date, datetime))

    self.configureLauncher(save_method=test_method)
    self.generatePromiseScript(promise_name, success=True)
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    self.assertTrue(os.path.exists(state_file))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))

    self.assertSuccessHistoryResult("my_promise")
    self.assertSuccessStatsResult(1)

  def test_runpromise_savemethod_no_logdir(self):
    promise_name = 'my_promise.py'
    def test_method(result):
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, TestResult))
      self.assertTrue(result.execution_time != 0)
      self.assertEqual(result.title, 'my_promise')
      self.assertEqual(result.name, promise_name)
      self.assertEqual(result.path, os.path.join(self.plugin_dir, promise_name))
      self.assertEqual(result.item.message, "success")
      self.assertEqual(result.item.hasFailed(), False)
      self.assertTrue(isinstance(result.item.date, datetime))

    # no promise log output dir
    self.configureLauncher(logdir=False, save_method=test_method)
    self.generatePromiseScript(promise_name, success=True)
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    self.assertTrue(os.path.exists(state_file))
    self.assertFalse(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))
    self.assertSuccessHistoryResult("my_promise")
    self.assertSuccessStatsResult(1)


  def test_runpromise_savemethod_anomaly(self):
    promise_name = 'my_promise.py'
    def test_method(result):
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, AnomalyResult))
      self.assertTrue(result.execution_time != 0)
      self.assertEqual(result.title, 'my_promise')
      self.assertEqual(result.name, promise_name)
      self.assertEqual(result.path, os.path.join(self.plugin_dir, promise_name))
      self.assertEqual(result.item.message, "success")
      self.assertEqual(result.item.hasFailed(), False)
      self.assertTrue(isinstance(result.item.date, datetime))

    self.configureLauncher(save_method=test_method, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True)
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    self.assertTrue(os.path.exists(state_file))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))
    self.assertSuccessHistoryResult("my_promise")
    self.assertSuccessStatsResult(1)


  def test_runpromise_savemethod_multiple(self):
    promise_name = 'my_promise.py'
    promise_failed = 'my_failed_promise.py'
    self.counter = 0
    def test_method(result):
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, TestResult))
      self.assertIn(result.name, [promise_failed, promise_name])
      if result.name == promise_failed:
        self.assertEqual(result.item.hasFailed(), True)
        self.assertEqual(result.item.message, "failed")
      else:
        self.assertEqual(result.item.hasFailed(), False)
        self.assertEqual(result.item.message, "success")
      self.counter += 1

    self.configureLauncher(save_method=test_method)
    self.generatePromiseScript(promise_name, success=True)
    self.generatePromiseScript(promise_failed, success=False)
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertEqual(self.counter, 2)
    self.assertTrue(os.path.exists(state_file))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_promise.log')))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_failed_promise.log')))

    self.assertSuccessHistoryResult("my_promise")
    self.assertSuccessHistoryResult("my_failed_promise", expected_history = """{
        "data": [{
                "failed": true,
                "message": "failed",
                "name": "%(name)s.py",
                "status": "ERROR",
                "title": "%(name)s"
        }]
}""")
    self.assertSuccessStatsResult(success=1, error=1)

  def test_runpromise_savemethod_multiple_success(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    third_promise = 'my_third_promise.py'
    self.counter = 0
    def test_method(result):
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, TestResult))
      self.assertIn(result.name, [first_promise, second_promise, third_promise])
      self.assertEqual(result.item.hasFailed(), False)
      self.assertEqual(result.item.message, "success")
      self.counter += 1

    self.configureLauncher(save_method=test_method)
    self.generatePromiseScript(first_promise, success=True)
    self.generatePromiseScript(second_promise, success=True)
    self.generatePromiseScript(third_promise, success=True)
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)

    # run promise will not fail
    self.launcher.run()
    self.assertEqual(self.counter, 3)
    self.assertTrue(os.path.exists(state_file))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_first_promise.log')))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_second_promise.log')))
    self.assertTrue(os.path.exists(os.path.join(self.log_dir, 'my_third_promise.log')))

    self.assertSuccessHistoryResult("my_first_promise")
    self.assertSuccessHistoryResult("my_second_promise")
    self.assertSuccessHistoryResult("my_third_promise")

    self.assertSuccessStatsResult(3)

  def test_runpromise_fail_and_success(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'

    self.configureLauncher()
    self.generatePromiseScript(first_promise, success=True)
    self.generatePromiseScript(second_promise, success=False)

    # run promise will fail when promise fail (usefull for slapgrid)
    with self.assertRaises(PromiseError) as exc:
      self.launcher.run()
    self.assertEqual(str(exc.exception), 'Promise %r failed with output: failed' % second_promise)

    # force to reload the module without rerun python
    os.system('rm %s/*.pyc' % self.plugin_dir)
    self.generatePromiseScript(second_promise, success=True)
    # wait next periodicity
    time.sleep(2)
    self.launcher.run()

    log_file = os.path.join(self.log_dir, 'my_second_promise.log')
    self.assertTrue(os.path.exists(log_file))
    with open(log_file) as f:
      line = f.readline()
      self.assertTrue('failed' in line, line)
      line = f.readline()
      self.assertTrue('success' in line, line)

    self.assertSuccessStatsResult(expected_stats = """{
        "data": ["Date, Success, Error, Warning",
                "__DATE__, 1, 1, ",
                "__DATE__, 2, 0, "
        ]
}""")

  def test_runpromise_with_periodicity(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    self.counter = 0

    def test_method_first(result):
      self.assertIn(result.name, [first_promise, second_promise])
      self.assertEqual(result.item.hasFailed(), False)
      self.assertEqual(result.item.message, "success")
      self.counter += 1

    def test_method_one(result):
      self.counter += 1
      self.assertEqual(result.name, first_promise)
      self.assertEqual(result.item.hasFailed(), False)
      self.assertEqual(result.item.message, "success")

    self.configureLauncher(save_method=test_method_first)
    # ~2 seconds
    self.generatePromiseScript(first_promise, success=True, periodicity=0.03)
    # ~3 seconds
    self.generatePromiseScript(second_promise, success=True, periodicity=0.05)

    self.launcher.run()
    self.assertEqual(self.counter, 2)

    self.configureLauncher(save_method=test_method_one)
    time.sleep(2)
    self.counter = 0
    self.launcher.run() # only my_first_promise will run
    self.assertEqual(self.counter, 1)

    time.sleep(3)
    self.counter = 0
    self.configureLauncher(save_method=test_method_first)
    self.launcher.run()
    self.assertEqual(self.counter, 2)

  def test_runpromise_with_periodicity_same(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    self.counter = 0

    def test_method(result):
      self.assertIn(result.name, [first_promise, second_promise])
      self.assertEqual(result.item.hasFailed(), False)
      self.assertEqual(result.item.message, "success")
      self.counter += 1

    self.configureLauncher(save_method=test_method)
    # ~2 seconds
    self.generatePromiseScript(first_promise, success=True, periodicity=0.03)
    self.generatePromiseScript(second_promise, success=True, periodicity=0.03)

    self.launcher.run()
    self.assertEqual(self.counter, 2)

    self.configureLauncher(save_method=test_method)
    time.sleep(1)
    self.counter = 0
    self.launcher.run() # run nothing
    self.assertEqual(self.counter, 0)

    time.sleep(1)
    self.counter = 0
    self.configureLauncher(save_method=test_method)
    self.launcher.run()
    self.assertEqual(self.counter, 2)

  def test_runpromise_with_periodicity_result_failed(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    first_state_file = os.path.join(self.partition_dir, PROMISE_RESULT_FOLDER_NAME,
                                    'my_first_promise.status.json')
    second_state_file = os.path.join(self.partition_dir, PROMISE_RESULT_FOLDER_NAME,
                                     'my_second_promise.status.json')

    self.configureLauncher()
    # ~2 seconds
    self.generatePromiseScript(first_promise, success=True, periodicity=0.03)
    # ~3 seconds
    self.generatePromiseScript(second_promise, success=False, periodicity=0.05)

    with self.assertRaises(PromiseError) as exc:
      self.launcher.run()
    self.assertEqual(str(exc.exception), 'Promise %r failed with output: failed' % second_promise)

    self.assertTrue(os.path.exists(first_state_file))
    self.assertTrue(os.path.exists(second_state_file))
    with open(first_state_file) as f:
      first_result = json.load(f)
    with open(second_state_file) as f:
      second_result = json.load(f)
    self.assertEqual(first_result['name'], first_promise)
    self.assertEqual(second_result['name'], second_promise)
    first_date = first_result['result']['date']
    second_date = second_result['result']['date']

    self.configureLauncher()
    time.sleep(2)
    with self.assertRaises(PromiseError) as exc:
      self.launcher.run() # only my_first_promise will run but second_promise still failing
    self.assertEqual(str(exc.exception), 'Promise %r failed with output: failed' % second_promise)

    with open(first_state_file) as f:
      first_result = json.load(f)
    with open(second_state_file) as f:
      second_result = json.load(f)
    self.assertNotEqual(first_result['result']['date'], first_date)
    self.assertEqual(second_result['result']['date'], second_date)
    first_date = first_result['result']['date']

    time.sleep(3)
    self.configureLauncher()
    with self.assertRaises(PromiseError) as exc:
      self.launcher.run()
    self.assertEqual(str(exc.exception), 'Promise %r failed with output: failed' % second_promise)

    with open(first_state_file) as f:
      first_result = json.load(f)
    with open(second_state_file) as f:
      second_result = json.load(f)
    self.assertNotEqual(first_result['result']['date'], first_date)
    self.assertNotEqual(second_result['result']['date'], second_date)

  def test_runpromise_with_periodicity_result_failed_and_ok(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    first_state_file = os.path.join(self.partition_dir, PROMISE_RESULT_FOLDER_NAME,
                                    'my_first_promise.status.json')
    second_state_file = os.path.join(self.partition_dir, PROMISE_RESULT_FOLDER_NAME,
                                     'my_second_promise.status.json')

    self.configureLauncher()
    # ~2 seconds
    self.generatePromiseScript(first_promise, success=True, periodicity=0.03)
    # ~3 seconds
    self.generatePromiseScript(second_promise, success=False, periodicity=0.05)

    with self.assertRaises(PromiseError) as exc:
      self.launcher.run()
    self.assertEqual(str(exc.exception), 'Promise %r failed with output: failed' % second_promise)

    self.assertTrue(os.path.exists(first_state_file))
    self.assertTrue(os.path.exists(second_state_file))
    with open(first_state_file) as f:
      first_result = json.load(f)
    with open(second_state_file) as f:
      second_result = json.load(f)
    self.assertEqual(first_result['name'], first_promise)
    self.assertEqual(second_result['name'], second_promise)
    first_date = first_result['result']['date']
    second_date = second_result['result']['date']

    self.configureLauncher()
    time.sleep(2)
    with self.assertRaises(PromiseError) as exc:
      self.launcher.run() # only my_first_promise will run but second_promise still failing
    self.assertEqual(str(exc.exception), 'Promise %r failed with output: failed' % second_promise)

    with open(first_state_file) as f:
      first_result = json.load(f)
    with open(second_state_file) as f:
      second_result = json.load(f)
    self.assertNotEqual(first_result['result']['date'], first_date)
    self.assertEqual(second_result['result']['date'], second_date)
    first_date = first_result['result']['date']
    second_date = second_result['result']['date']

    time.sleep(4)
    # force to reload the module without rerun python
    os.system('rm %s/*.pyc' % self.plugin_dir)

    # second_promise is now success
    self.generatePromiseScript(second_promise, success=True, periodicity=0.05)
    self.configureLauncher()
    self.launcher.run() # now all succeed

    with open(first_state_file) as f:
      first_result = json.load(f)
    with open(second_state_file) as f:
      second_result = json.load(f)
    self.assertNotEqual(first_result['result']['date'], first_date)
    self.assertNotEqual(second_result['result']['date'], second_date)

  def test_runpromise_force(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    self.counter = 0

    def test_method(result):
      self.assertIn(result.name, [first_promise, second_promise])
      self.assertEqual(result.item.hasFailed(), False)
      self.assertEqual(result.item.message, "success")
      self.counter += 1

    self.configureLauncher(save_method=test_method)
    # ~2 seconds
    self.generatePromiseScript(first_promise, success=True, periodicity=0.03)
    self.generatePromiseScript(second_promise, success=True, periodicity=0.03)

    self.launcher.run()
    self.assertEqual(self.counter, 2)

    self.configureLauncher(save_method=test_method)
    time.sleep(1)
    self.counter = 0
    self.launcher.run() # run nothing
    self.assertEqual(self.counter, 0)

    self.configureLauncher(save_method=test_method, force=True)
    self.counter = 0
    self.launcher.run() # will run all as force is True
    self.assertEqual(self.counter, 2)

    self.configureLauncher(save_method=test_method)
    time.sleep(1)
    self.counter = 0
    self.launcher.run() # run nothing
    self.assertEqual(self.counter, 0)

    time.sleep(1)
    self.counter = 0
    self.configureLauncher(save_method=test_method)
    self.launcher.run() # after 2 seconds will run all
    self.assertEqual(self.counter, 2)

  def test_runpromise_wrapped(self):
    promise_name = "my_bash_promise"
    promise_path = os.path.join(self.legacy_promise_dir, promise_name)
    self.called = False
    with open(promise_path, 'w') as f:
      f.write("""#!/bin/bash
echo "success"
""")
    os.chmod(promise_path, 0o744)

    def test_method(result):
      self.called = True
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, TestResult))
      self.assertTrue(result.execution_time != 0)
      self.assertEqual(result.title, promise_name)
      self.assertEqual(result.name, promise_name)
      self.assertEqual(result.path, os.path.join(self.legacy_promise_dir, promise_name))
      self.assertEqual(result.item.message, "success")
      self.assertEqual(result.item.hasFailed(), False)
      self.assertTrue(isinstance(result.item.date, datetime))

    self.configureLauncher(save_method=test_method)
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)
    self.launcher.run()

    self.assertTrue(self.called)
    self.assertTrue(os.path.exists(state_file))

  def test_runpromise_wrapped_failed(self):
    promise_name = "my_bash_promise"
    promise_path = os.path.join(self.legacy_promise_dir, promise_name)
    with open(promise_path, 'w') as f:
      f.write("""#!/bin/bash
echo "This promise failed"
exit 1
""")
    os.chmod(promise_path, 0o744)

    self.configureLauncher()
    state_file = os.path.join(self.partition_dir, PROMISE_STATE_FOLDER_NAME)
    with self.assertRaises(PromiseError) as exc:
      self.launcher.run()
    self.assertEqual(str(exc.exception), 'Promise %r failed with output: This promise failed' % promise_name)

  def test_runpromise_wrapped_mixed(self):
    self.called = 0
    result_dict = {"my_bash_promise": "", "my_bash_promise2": "", "first_promise.py": "", "second_promise.py": ""}
    def test_method(result):
      self.called += 1
      result_dict.pop(result.name)
      if result.title == "first_promise" or result.title == "second_promise":
        self.assertEqual(result.item.message, "success")
      if result.title == "my_bash_promise":
        self.assertEqual(result.item.message, "promise 1 succeeded")
      if result.title == "my_bash_promise2":
        self.assertEqual(result.item.message, "promise 2 succeeded")
      self.assertEqual(result.item.hasFailed(), False)

    promise_name = "my_bash_promise"
    promise_path = os.path.join(self.legacy_promise_dir, promise_name)
    promise_name2 = "my_bash_promise2"
    promise_path2 = os.path.join(self.legacy_promise_dir, promise_name2)
    with open(promise_path, 'w') as f:
      f.write("""#!/bin/bash
echo "promise 1 succeeded"
exit 0
""")
    os.chmod(promise_path, 0o744)
    with open(promise_path2, 'w') as f:
      f.write("""#!/bin/bash
echo "promise 2 succeeded"
exit 0
""")
    os.chmod(promise_path2, 0o744)
    self.generatePromiseScript("first_promise.py", success=True)
    self.generatePromiseScript("second_promise.py", success=True)

    self.configureLauncher(save_method=test_method)
    self.launcher.run()
    self.assertEqual(self.called, 4)


  def test_runpromise_run_only(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    third_promise = 'my_third_promise.py'
    self.counter = 0
    self.check_list = [first_promise, second_promise, third_promise]
    def test_method(result):
      self.assertIn(result.name, self.check_list)
      self.assertEqual(result.item.hasFailed(), False)
      self.assertEqual(result.item.message, "success")
      self.counter += 1

    self.configureLauncher(save_method=test_method)
    self.generatePromiseScript(first_promise, success=True)
    self.generatePromiseScript(second_promise, success=True)
    self.generatePromiseScript(third_promise, success=True)

    # run promise will not fail
    self.launcher.run()
    self.assertEqual(self.counter, 3)

    self.counter = 0
    self.check_list = [second_promise]
    self.configureLauncher(save_method=test_method, run_list=[second_promise], force=True)
    time.sleep(1)
    self.launcher.run()
    self.assertEqual(self.counter, 1)

  def test_runpromise_run_only_multiple(self):
    first_promise = 'my_first_promise.py'
    second_promise = 'my_second_promise.py'
    third_promise = 'my_third_promise.py'
    self.counter = 0
    self.check_list = [first_promise, second_promise, third_promise]
    def test_method(result):
      self.assertIn(result.name, self.check_list)
      self.assertEqual(result.item.hasFailed(), False)
      self.assertEqual(result.item.message, "success")
      self.counter += 1

    self.configureLauncher(save_method=test_method)
    self.generatePromiseScript(first_promise, success=True)
    self.generatePromiseScript(second_promise, success=True)
    self.generatePromiseScript(third_promise, success=True)

    # run promise will not fail
    self.launcher.run()
    self.assertEqual(self.counter, 3)

    self.counter = 0
    self.check_list = [third_promise, second_promise]
    self.configureLauncher(save_method=test_method, run_list=self.check_list, force=True)
    time.sleep(1)
    self.launcher.run()
    self.assertEqual(self.counter, 2)

  def writeLatestPromiseResult(self):
    state_file = os.path.join(self.partition_dir, PROMISE_RESULT_FOLDER_NAME, 'my_promise.status.json')
    result_string = """{
  "result": {
    "failed": true,
    "message": "error",
    "date": "%s",
    "type": "Test Result"
  },
  "path": "%s/my_promise.py",
  "name": "my_promise.py",
  "execution-time": 0.05,
  "title": "my_promise"
}""" % (datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S+0000'), self.plugin_dir)

    if not os.path.exists(os.path.dirname(state_file)):
      os.makedirs(os.path.dirname(state_file))
    with open(state_file, 'w') as f:
      f.write(result_string)

  def writeLatestBashPromiseResult(self, name="my_bash_promise"):
    state_file = os.path.join(self.partition_dir, PROMISE_RESULT_FOLDER_NAME,
                              '%s.status.json' % name)
    result_string = """{
  "result": {
    "failed": true,
    "message": "error",
    "date": "%(date)s",
    "type": "Test Result"
  },
  "path": "%(folder)s/%(name)s",
  "name": "%(name)s",
  "execution-time": 0.05,
  "title": "%(name)s"
}""" % {'date': datetime.utcnow().strftime('%Y-%m-%dT%H:%M:%S+0000'),
        'folder': self.legacy_promise_dir,
        'name': name}

    if not os.path.exists(os.path.dirname(state_file)):
      os.makedirs(os.path.dirname(state_file))
    with open(state_file, 'w') as f:
      f.write(result_string)

  def test_runpromise_will_timeout(self):
    self.called = False
    promise_name = 'my_promise.py'
    self.writeLatestPromiseResult()

    def test_method(result):
      self.called = True
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, AnomalyResult))
      self.assertTrue(result.execution_time >= 1)
      self.assertEqual(result.title, 'my_promise')
      self.assertEqual(result.name, promise_name)
      self.assertIn("Promise timed out after", result.item.message)
      self.assertEqual(result.item.hasFailed(), True)

    self.configureLauncher(save_method=test_method, enable_anomaly=True, timeout=1)
    self.generatePromiseScript(promise_name, success=True, content="""import time
    time.sleep(20)""")

    # run promise will timeout
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertTrue(self.called)

  def test_runpromise_wrapped_will_timeout(self):
    promise_name = "my_bash_promise"
    promise_path = os.path.join(self.legacy_promise_dir, promise_name)
    self.writeLatestBashPromiseResult()
    self.called = False
    with open(promise_path, 'w') as f:
      f.write("""#!/bin/bash
sleep 20
echo "success"
""")
    os.chmod(promise_path, 0o744)

    def test_method(result):
      self.called = True
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, TestResult))
      self.assertTrue(result.execution_time >= 1)
      self.assertEqual(result.title, promise_name)
      self.assertEqual(result.name, promise_name)
      self.assertEqual(result.path, promise_path)
      self.assertIn("Promise timed out after", result.item.message)
      self.assertEqual(result.item.hasFailed(), True)
      self.assertTrue(isinstance(result.item.date, datetime))

    self.configureLauncher(save_method=test_method, timeout=1)
    # run promise will timeout
    with self.assertRaises(PromiseError):
      self.launcher.run()
    self.assertTrue(self.called)

  def test_runpromise_wrapped_will_timeout_two(self):
    first_promise = "my_bash_promise"
    first_promise_path = os.path.join(self.legacy_promise_dir, first_promise)
    second_promise = "my_second_bash_promise"
    second_promise_path = os.path.join(self.legacy_promise_dir, second_promise)
    self.writeLatestBashPromiseResult(first_promise)
    self.writeLatestBashPromiseResult(second_promise)

    def createPromise(promise_path):
      with open(promise_path, 'w') as f:
        f.write("""#!/bin/bash
echo "some data from promise"
sleep 20
echo "success"
exit 1
""")
      os.chmod(promise_path, 0o744)

    createPromise(first_promise_path)
    createPromise(second_promise_path)
    self.configureLauncher(timeout=0.5)
    # run promise will timeout
    with self.assertRaises(PromiseError):
      self.launcher.run()

  def test_runpromise_timeout_fail_if_test(self):
    promise_name = 'my_promise.py'

    self.configureLauncher(timeout=1, enable_anomaly=False)
    self.generatePromiseScript(promise_name, success=True, content="""import time
    time.sleep(20)""", periodicity=0.01)

    # timeout for the first time and raise
    with self.assertRaises(PromiseError):
      self.launcher.run()

  def test_runpromise_fail_if_timeout_twice(self):
    promise_name = 'my_promise.py'

    self.configureLauncher(timeout=1, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True, content="""import time
    time.sleep(20)""", periodicity=0.01)

    # timeout for the first time, no raise
    self.launcher.run()

    # run promise will timeout and raise
    time.sleep(1)
    with self.assertRaises(PromiseError):
      self.launcher.run()

    # run promise will continue to raise
    time.sleep(1)
    with self.assertRaises(PromiseError):
      self.launcher.run()

  def test_runpromise_not_tested(self):
    promise_name = 'my_promise.py'

    def test_method(result):
      self.called = True

    self.called = False
    self.configureLauncher(save_method=test_method, timeout=5, enable_anomaly=False)
    self.generatePromiseScript(promise_name, success=True, content="""import time
    time.sleep(20)""", periodicity=0.01, is_tested=False,)

    # will not run the promise in test mode (so no sleep)
    self.launcher.run()
    # no result returned by the promise
    self.assertFalse(self.called)

  def test_runpromise_not_tested_with_anomaly(self):
    promise_name = 'my_promise.py'

    def test_method(result):
      self.called = True
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, AnomalyResult))

    self.called = False
    self.configureLauncher(save_method=test_method, timeout=1, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True,
      periodicity=0.01, is_tested=False,)

    # will run the promise because we are in anomaly mode
    self.launcher.run()
    # promise result is saved
    self.assertTrue(self.called)

  def test_runpromise_without_anomaly(self):
    promise_name = 'my_promise.py'

    def test_method(result):
      self.called = True

    self.called = False
    self.configureLauncher(save_method=test_method, timeout=1, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True, content="""import time
    time.sleep(20)""", periodicity=0.01, with_anomaly=False,)

    # will not run the promise in anomaly mode (so no sleep)
    self.launcher.run()
    # no result returned by the promise
    self.assertFalse(self.called)

  def test_runpromise_without_anomaly_but_test(self):
    promise_name = 'my_promise.py'

    def test_method(result):
      self.called = True
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, TestResult))

    self.called = False
    self.configureLauncher(save_method=test_method, timeout=1, enable_anomaly=False)
    self.generatePromiseScript(promise_name, success=True,
      periodicity=0.01, with_anomaly=False,)

    # will run the promise because we are in anomaly mode
    self.launcher.run()
    # promise result is saved
    self.assertTrue(self.called)

  def test_runpromise_not_tested_will_not_change_periodicity(self):
    promise_name = 'my_promise.py'

    def test_method(result):
      self.called = True

    self.called = False

    self.configureLauncher(save_method=test_method, timeout=5, enable_anomaly=False)
    self.generatePromiseScript(promise_name, success=True, periodicity=5,
      is_tested=False,)

    # will not run the promise in test mode
    self.launcher.run()
    self.assertFalse(self.called)

    self.configureLauncher(save_method=test_method, timeout=5, enable_anomaly=True)
    # will not run immediately anomaly
    self.launcher.run()
    # no result returned by the promise
    self.assertTrue(self.called)

  def test_runpromise_without_anomaly_no_change_periodicity(self):
    promise_name = 'my_promise.py'

    def test_method(result):
      self.called = True

    self.called = False

    self.configureLauncher(save_method=test_method, timeout=5, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True, periodicity=5,
      with_anomaly=False,)

    # will not run the promise in anomaly mode
    self.launcher.run()
    self.assertFalse(self.called)

    self.configureLauncher(save_method=test_method, timeout=5, enable_anomaly=False)
    # will not run immediately anomaly (periodicity didn't change)
    self.launcher.run()
    # no result returned by the promise
    self.assertTrue(self.called)

  def test_runpromise_not_tested_without_anomaly_fail(self):
    promise_name = 'my_promise.py'

    def test_method(result):
      self.called = True
      self.assertTrue(isinstance(result, PromiseQueueResult))
      self.assertTrue(isinstance(result.item, AnomalyResult))
      self.assertEqual(result.item.message, "It's not allowed to disable both Test and Anomaly in promise!")
      self.assertEqual(result.item.hasFailed(), True)

    self.called = False
    self.configureLauncher(save_method=test_method, timeout=1, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True, content="""import time
    time.sleep(20)""", periodicity=0.01, with_anomaly=False, is_tested=False)

    # will fail because disable anomaly and test is not allowed
    with self.assertRaises(PromiseError):
      self.launcher.run()
    # no result returned by the promise
    self.assertTrue(self.called)

  def test_promise_cache(self):
    promise_name = 'my_promise.py'
    promise_file = os.path.join(self.plugin_dir, promise_name)
    self.configureLauncher(timeout=1, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True, periodicity=0.01,
      with_anomaly=True, is_tested=False)

    # run promise, no failure
    self.launcher.run()
    cache_folder = os.path.join(self.partition_dir, PROMISE_CACHE_FOLDER_NAME)
    cache_file = os.path.join(cache_folder, 'my_promise')
    self.assertTrue(os.path.exists(cache_folder))
    self.assertTrue(os.path.exists(cache_file))
    file_stat = os.stat(promise_file)
    with open(cache_file) as f:
      cache_dict = json.load(f)
    timestamp = cache_dict.pop('timestamp')
    info_dict = {
      u'is_tested': False,
      u'is_anomaly_detected': True,
      u'periodicity': 0.01,
      u'next_run_after' : (timestamp + 0.01 * 60.0),
      u'module_file': u'%s' % promise_file,
      u'module_file_mtime': file_stat.st_mtime,
    }
    # next run is in future
    self.assertTrue(info_dict['next_run_after'] > time.time())
    self.assertEqual(info_dict, cache_dict)

  def test_promise_cache_expire_with_periodicity(self):
    self.called = False
    def test_method(result):
      self.called = True

    promise_name = 'my_promise.py'
    promise_file = os.path.join(self.plugin_dir, promise_name)
    self.configureLauncher(save_method=test_method, timeout=1, enable_anomaly=True)
    self.generatePromiseScript(promise_name, success=True, periodicity=0.01,
      with_anomaly=True, is_tested=False)

    # run promise, no failure
    self.launcher.run()
    cache_folder = os.path.join(self.partition_dir, PROMISE_CACHE_FOLDER_NAME)
    cache_file = os.path.join(cache_folder, 'my_promise')
    self.assertTrue(os.path.exists(cache_folder))
    self.assertTrue(os.path.exists(cache_file))
    file_stat = os.stat(promise_file)
    with open(cache_file) as f:
      cache_dict = json.load(f)
    timestamp = cache_dict.pop('timestamp')
    info_dict = {
      u'is_tested': False,
      u'is_anomaly_detected': True,
      u'periodicity': 0.01,
      u'next_run_after' : (timestamp + 0.01 * 60.0),
      u'module_file': u'%s' % promise_file,
      u'module_file_mtime': file_stat.st_mtime,
    }
    self.assertEqual(info_dict, cache_dict)
    self.assertTrue(self.called)
    next_run_after = cache_dict['next_run_after']

    # periodicity not match
    self.called = False
    self.configureLauncher(save_method=test_method, timeout=1, enable_anomaly=True)
    self.launcher.run()
    self.assertFalse(self.called)
    with open(cache_file) as f:
      cache_dict = json.load(f)
    # no change!
    current_timestamp = cache_dict.pop('timestamp')
    self.assertEqual(current_timestamp, timestamp)
    self.assertEqual(info_dict, cache_dict)

    time.sleep(1)
    # periodicity match
    self.configureLauncher(save_method=test_method, timeout=1, enable_anomaly=True)
    self.launcher.run()

    # cached was updated
    with open(cache_file) as f:
      cache_dict = json.load(f)
    new_timestamp = cache_dict.pop('timestamp')
    info_dict = {
      u'is_tested': False,
      u'is_anomaly_detected': True,
      u'periodicity': 0.01,
      u'next_run_after' : (new_timestamp + 0.01 * 60.0),
      u'module_file': u'%s' % promise_file,
      u'module_file_mtime': file_stat.st_mtime,
    }
    self.assertTrue(new_timestamp > timestamp)
    # next run is in future
    self.assertTrue(cache_dict['next_run_after'] > next_run_after)
    self.assertEqual(info_dict, cache_dict)

class TestSlapOSGenericPromise(TestSlapOSPromiseMixin):

  def initialisePromise(self, promise_content="", success=True, timeout=60):
    self.promise_name = 'my_promise.py'
    self.promise_path = os.path.join(self.plugin_dir, self.promise_name)
    self.configureLauncher()
    self.generatePromiseScript(self.promise_name, periodicity=1, content=promise_content, success=success)
    self.writeInit()
    self.queue = queue.Queue()
    self.promise_config = {
      'log-folder': self.log_dir,
      'partition-folder': self.partition_dir,
      'promise-timeout': timeout,
      'debug': False,
      'check-anomaly': True,
      'master-url': "https://master.url.com",
      'partition-cert': '',
      'partition-key': '',
      'partition-id': self.partition_id,
      'computer-id': self.computer_id,
      'queue': self.queue,
      'path': self.promise_path,
      'name': self.promise_name
    }

  def createPromiseProcess(self, check_anomaly=False, wrap=False):
    return PromiseProcess(
      self.partition_dir,
      self.promise_name,
      self.promise_path,
      logger=logging.getLogger('slapos.test.promise'),
      argument_dict=self.promise_config,
      check_anomaly=check_anomaly,
      wrap=wrap,
    )

  def test_create_simple_promise(self):
    self.initialisePromise()
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)
    self.assertEqual(promise.getPeriodicity(), 1)
    self.assertEqual(promise.getName(), self.promise_name)
    self.assertEqual(promise.getTitle(), 'my_promise')
    self.assertEqual(promise.getPartitionFolder(), self.partition_dir)
    self.assertEqual(promise.getPromiseFile(), self.promise_path)
    self.assertEqual(promise.getLogFolder(), self.log_dir)
    self.assertEqual(promise.getLogFile(), os.path.join(self.log_dir, 'my_promise.log'))

    promise.setPeriodicity(2)
    self.assertEqual(promise.getPeriodicity(), 2)
    with self.assertRaises(ValueError):
      promise.setPeriodicity(0)

    promise.run(check_anomaly=True)
    result = self.queue.get(True, 1)
    self.assertTrue(isinstance(result, PromiseQueueResult))
    self.assertTrue(isinstance(result.item, AnomalyResult))
    self.assertEqual(result.title, 'my_promise')
    self.assertEqual(result.name, self.promise_name)
    self.assertEqual(result.path, os.path.join(self.plugin_dir, self.promise_name))
    self.assertEqual(result.item.message, "success")
    self.assertEqual(result.item.hasFailed(), False)
    self.assertTrue(isinstance(result.item.date, datetime))

  def test_promise_cleanup_plugin_dir(self):
    stale_pyc = os.path.join(self.plugin_dir, 'stale.pyc')
    with open(stale_pyc, 'w') as fh:
      fh.write('')
    stale_pyo = os.path.join(self.plugin_dir, 'stale.pyo')
    with open(stale_pyo, 'w') as fh:
      fh.write('')
    self.initialisePromise()
    self.launcher.run()
    self.assertFalse(os.path.exists(stale_pyc))
    self.assertFalse(os.path.exists(stale_pyo))

  def test_promise_cleanup_output_history(self):
    promise_name = 'dummy.py'
    promise_content = """from zope.interface import implementer
from slapos.grid.promise import interface
from slapos.grid.promise import GenericPromise

@implementer(interface.IPromise)
class RunPromise(GenericPromise):

  def sense(self):
    self.logger.info("success")

  def anomaly(self):
    return self._anomaly()

  def test(self):
    return self._test()
"""

    promise_path = os.path.join(self.plugin_dir, promise_name)
    self.writeFile(promise_path, promise_content)
    self.writeInit()

    # output .slapgrid/promise/result
    # history .slapgrid/promise/history'
    promise_dir = os.path.join(self.partition_dir, '.slapgrid', 'promise')
    result_dir = os.path.join(promise_dir, 'result')
    history_dir = os.path.join(promise_dir, 'history')
    os.makedirs(result_dir)
    os.makedirs(history_dir)

    def createFile(path, name, content=''):
      filepath = os.path.join(path, name)
      with open(filepath, 'w') as fh:
        fh.write(content)
      return filepath

    promise_status_json  = createFile(
      promise_dir, 'promise_status.json',
      '{"stale.py": ["OK", "2020-01-09T08:09:44+0000", '
      '"260ca9dd8a4577fc00b7bd5810298076"]}')

    dummy_result = createFile(result_dir, 'dummy.status.json')
    dummy_history = createFile(history_dir, 'dummy.history.json')
    dummy_history_old = createFile(history_dir, 'dummy.history.old.json')

    stale_result = createFile(result_dir, 'stale.status.json')
    stale_history = createFile(history_dir, 'stale.history.json')
    stale_history_old = createFile(history_dir, 'stale.history.old.json')

    just_result_file = createFile(result_dir, 'doesnotmatch')
    just_history_file = createFile(history_dir, 'doesnotmatch')

    def assertPathExists(path):
      self.assertTrue(os.path.exists(path))

    def assertPathNotExists(path):
      self.assertFalse(os.path.exists(path))

    self.initialisePromise()
    self.launcher.run()

    assertPathExists(dummy_result)
    assertPathExists(dummy_history)
    assertPathExists(dummy_history_old)

    # check that it does not clean up too much
    assertPathExists(just_result_file)
    assertPathExists(just_history_file)

    assertPathNotExists(stale_result)
    assertPathNotExists(stale_history)
    assertPathNotExists(stale_history_old)

    with open(promise_status_json) as fh:
      promise_status = json.load(fh)
    self.assertNotIn('stale.py', promise_status)

  def test_promise_anomaly_disabled(self):
    self.initialisePromise()
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    promise.run()
    result = self.queue.get(True, 1)
    self.assertTrue(isinstance(result, PromiseQueueResult))
    self.assertTrue(isinstance(result.item, TestResult))
    self.assertEqual(result.title, 'my_promise')
    self.assertEqual(result.name, self.promise_name)
    self.assertEqual(result.path, os.path.join(self.plugin_dir, self.promise_name))
    self.assertEqual(result.item.message, "success")
    self.assertEqual(result.item.hasFailed(), False)
    self.assertTrue(isinstance(result.item.date, datetime))

  def test_promise_with_raise(self):
    promise_content = "raise ValueError('Bad Promise raised')"
    self.initialisePromise(promise_content)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    # no raise
    promise.run()
    result = self.queue.get(True, 1)
    self.assertEqual(result.title, 'my_promise')
    self.assertEqual(result.name, self.promise_name)
    self.assertEqual(result.item.message, "Bad Promise raised")
    self.assertEqual(result.item.hasFailed(), True)

  def test_promise_no_return(self):
    promise_content = "return"
    self.initialisePromise(promise_content)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    # no raise
    promise.run()
    result = self.queue.get(True, 1)
    self.assertEqual(result.title, 'my_promise')
    self.assertEqual(result.name, self.promise_name)
    self.assertEqual(result.item.message, "No result found!")
    self.assertEqual(result.item.hasFailed(), False)

  def test_promise_resultfromlog(self):
    promise_content = "self.logger.info('Promise is running...')"
    self.initialisePromise(promise_content)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    date = datetime.now()
    promise.sense()
    # get all messages during the latest minute
    latest_message_list = promise.getLastPromiseResultList(result_count=1)
    date = datetime.strptime(date.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S')
    self.assertEqual(len(latest_message_list), 1)
    self.assertEqual(
      latest_message_list[0][0],
      {'date': date, 'status': 'INFO', 'message': 'Promise is running...'})
    self.assertEqual(
      latest_message_list[0][1],
      {'date': date, 'status': 'INFO', 'message': 'success'})

  def test_promise_resultfromlog_error(self):
    promise_content = 'self.logger.error("Promise is running...\\nmessage in new line")'
    self.initialisePromise(promise_content)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    date = datetime.now()
    promise.sense()
    # get all messages during the latest minute
    latest_message_list = promise.getLastPromiseResultList(result_count=1)
    date = datetime.strptime(date.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S')
    self.assertEqual(len(latest_message_list), 1)
    self.assertEqual(
      latest_message_list[0][0],
      {'date': date, 'status': 'ERROR',
       'message': 'Promise is running...\nmessage in new line'})
    self.assertEqual(
      latest_message_list[0][1],
      {'date': date, 'status': 'INFO', 'message': 'success'})

  def test_promise_resultfromlog_no_logfolder(self):
    self.log_dir = None
    promise_content = "self.logger.info('Promise is running...')"
    self.initialisePromise(promise_content)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    date = datetime.now()
    promise.sense()
    self.assertEqual(promise.getLogFolder(), None)
    self.assertEqual(promise.getLogFile(), None)

    # get all messages during the latest minute
    latest_message_list = promise.getLastPromiseResultList(result_count=1)
    date = datetime.strptime(date.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S')
    self.assertEqual(len(latest_message_list), 1)
    self.assertEqual(
      latest_message_list[0][0],
      {'date': date, 'status': 'INFO', 'message': 'Promise is running...'})
    self.assertEqual(
      latest_message_list[0][1],
      {'date': date, 'status': 'INFO', 'message': 'success'})

  def test_promise_resultfromlog_latest_minutes(self):
    self.initialisePromise(timeout=60)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    # write some random logs
    start_date = datetime.now()
    with open(promise.getLogFile(), 'w') as f:
      for i in range(0, 50):
        transaction_id = '%s-%s' % (int(time.time()), random.randint(100, 999))
        date = start_date - timedelta(minutes=(49 - i))
        date_string = date.strftime('%Y-%m-%d %H:%M:%S')
        line = "%s - INFO - %s - Promise result %s\n" % (date_string, transaction_id, i)
        f.write(line)

    latest_message_list = promise.getLastPromiseResultList(
      latest_minute=10,
      result_count=100
    )
    start_date = datetime.strptime(start_date.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S')
    end_date_string = (start_date - timedelta(minutes=9)).strftime('%Y-%m-%d %H:%M:%S')
    end_date = datetime.strptime(end_date_string, '%Y-%m-%d %H:%M:%S')
    self.assertEqual(len(latest_message_list), 10)
    for message in latest_message_list:
      self.assertEqual(len(message), 1)
    self.assertEqual(
      latest_message_list[0][0],
      {'date': start_date, 'status': 'INFO', 'message': 'Promise result 49'})
    self.assertEqual(
      latest_message_list[-1][0],
      {'date': end_date, 'status': 'INFO', 'message': 'Promise result 40'})

  def test_promise_resultfromlog_latest_minutes_multilog(self):
    self.initialisePromise(timeout=60)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    # write some random logs
    start_date = datetime.now()
    date = start_date
    line_list = []
    j = 0
    for i in range(0, 25):
      transaction_id = '%s-%s' % (int(time.time()), random.randint(100, 999))
      date_string = date.strftime('%Y-%m-%d %H:%M:%S')
      line_list.append("%s - INFO - %s - Promise result %s\n" % (date_string, transaction_id, j))
      date = date - timedelta(seconds=30)
      j += 1
      date_string = date.strftime('%Y-%m-%d %H:%M:%S')
      line_list.append("%s - INFO - %s - Promise result %s\n" % (date_string, transaction_id, j))
      date = date - timedelta(seconds=30)
      j += 1

    line_list.reverse()
    with open(promise.getLogFile(), 'w') as f:
      f.write('\n'.join(line_list))

    latest_message_list = promise.getLastPromiseResultList(
      latest_minute=10,
      result_count=100
    )

    start_date = datetime.strptime(start_date.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S')
    end_date_string = (start_date - timedelta(seconds=30*19)).strftime('%Y-%m-%d %H:%M:%S')
    end_date = datetime.strptime(end_date_string, '%Y-%m-%d %H:%M:%S')
    # there is 2 result line per minutes
    self.assertEqual(len(latest_message_list), 10)
    for message in latest_message_list:
      self.assertEqual(len(message), 2)
    self.assertEqual(
      latest_message_list[0][1],
      {'date': start_date, 'status': 'INFO', 'message': 'Promise result 0'})
    self.assertEqual(
      latest_message_list[-1][0],
      {'date': end_date, 'status': 'INFO', 'message': 'Promise result 19'})

  def test_promise_resultfromlog_result_count(self):
    self.initialisePromise()
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    # write some random logs
    start_date = datetime.now()
    date = start_date
    line_list = []
    j = 0
    for i in range(0, 25):
      transaction_id = '%s-%s' % (int(time.time()), random.randint(100, 999))
      date_string = date.strftime('%Y-%m-%d %H:%M:%S')
      line_list.append("%s - INFO - %s - Promise result %s\n" % (date_string, transaction_id, j))
      date = date - timedelta(seconds=30)
      j += 1
      date_string = date.strftime('%Y-%m-%d %H:%M:%S')
      line_list.append("%s - INFO - %s - Promise result %s\n" % (date_string, transaction_id, j))
      date = date - timedelta(seconds=30)
      j += 1

    line_list.reverse()
    with open(promise.getLogFile(), 'w') as f:
      f.write('\n'.join(line_list))

    # result_count = 2 will return 2 log
    # max execution time is 1 min and log is writen every 30 seconds
    latest_message_list = promise.getLastPromiseResultList(result_count=1)

    start_date = datetime.strptime(start_date.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S')
    end_date_string = (start_date - timedelta(seconds=30)).strftime('%Y-%m-%d %H:%M:%S')
    end_date = datetime.strptime(end_date_string, '%Y-%m-%d %H:%M:%S')
    # there is 2 result line per minutes
    self.assertEqual(len(latest_message_list), 1)
    self.assertEqual(
      latest_message_list[0][0],
      {'date': end_date, 'status': 'INFO', 'message': 'Promise result 1'})
    self.assertEqual(
      latest_message_list[0][1],
      {'date': start_date, 'status': 'INFO', 'message': 'Promise result 0'})

  def test_promise_resultfromlog_result_count_many(self):
    self.initialisePromise()
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    # write some random logs
    start_date = datetime.now()
    date = start_date
    line_list = []
    j = 0
    for i in range(0, 25):
      transaction_id = '%s-%s' % (int(time.time()), random.randint(100, 999))
      date_string = date.strftime('%Y-%m-%d %H:%M:%S')
      line_list.append("%s - INFO - %s - Promise result %s\n" % (date_string, transaction_id, j))
      date = date - timedelta(seconds=30)
      j += 1
      date_string = date.strftime('%Y-%m-%d %H:%M:%S')
      line_list.append("%s - INFO - %s - Promise result %s\n" % (date_string, transaction_id, j))
      date = date - timedelta(seconds=30)
      j += 1

    line_list.reverse()
    with open(promise.getLogFile(), 'w') as f:
      f.write('\n'.join(line_list))

    # result_count = 2 will return 4 log
    # max execution time is 1 min and log is writen every 30 seconds
    latest_message_list = promise.getLastPromiseResultList(result_count=2)

    start_date = datetime.strptime(start_date.strftime('%Y-%m-%d %H:%M:%S'), '%Y-%m-%d %H:%M:%S')
    end_date_string = (start_date - timedelta(seconds=30*3)).strftime('%Y-%m-%d %H:%M:%S')
    end_date = datetime.strptime(end_date_string, '%Y-%m-%d %H:%M:%S')
    # there is 2 result line per minutes
    self.assertEqual(len(latest_message_list), 2)
    for message in latest_message_list:
      self.assertEqual(len(message), 2)
    self.assertEqual(
      latest_message_list[0][1],
      {'date': start_date, 'status': 'INFO', 'message': 'Promise result 0'})
    self.assertEqual(
      latest_message_list[-1][0],
      {'date': end_date, 'status': 'INFO', 'message': 'Promise result 3'})

    latest_message_list = promise.getLastPromiseResultList(result_count=100)
    # all results
    self.assertEqual(len(latest_message_list), 25)

  def test_promise_defaulttest(self):
    promise_content = 'self.logger.info("Promise is running...\\nmessage in new line")'
    self.initialisePromise(promise_content)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    promise.sense()

    result = promise._test(result_count=1, failure_amount=1)
    self.assertTrue(isinstance(result, TestResult))
    self.assertEqual(result.message, 'Promise is running...\nmessage in new line\nsuccess')
    self.assertEqual(result.hasFailed(), False)

  def test_promise_defaulttest_failure(self):
    self.initialisePromise(success=False)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    promise.sense()

    result = promise._test(result_count=1, failure_amount=1)
    self.assertTrue(isinstance(result, TestResult))
    self.assertEqual(result.message, 'failed')
    self.assertEqual(result.hasFailed(), True)

  def test_promise_defaulttest_error_if_two_fail(self):
    self.initialisePromise(success=False, timeout=1)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    promise.sense()

    # fail if 2 errors found
    result = promise._test(result_count=2, failure_amount=2)
    self.assertTrue(isinstance(result, TestResult))
    self.assertEqual(result.message, 'failed')
    self.assertEqual(result.hasFailed(), False)

    self.initialisePromise(success=False, timeout=1)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)
    promise.sense()
    result = promise._test(result_count=2, failure_amount=2)
    self.assertEqual(result.message, 'failed')
    self.assertEqual(result.hasFailed(), True)

    # will continue to fail
    self.initialisePromise(success=False, timeout=1)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)
    promise.sense()
    result = promise._test(result_count=2, failure_amount=2)
    self.assertEqual(result.message, 'failed')
    self.assertEqual(result.hasFailed(), True)

  def test_promise_defaulttest_anomaly(self):
    promise_content = 'self.logger.info("Promise is running...\\nmessage in new line")'
    self.initialisePromise(promise_content)
    promise_process = self.createPromiseProcess()
    promise_module = promise_process._loadPromiseModule()
    promise = promise_module.RunPromise(self.promise_config)

    promise.sense()

    result = promise._anomaly(result_count=1, failure_amount=1)
    self.assertTrue(isinstance(result, AnomalyResult))
    self.assertEqual(result.message, 'Promise is running...\nmessage in new line\nsuccess')
    self.assertEqual(result.hasFailed(), False)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2018 Vifib SARL and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

# pyright: strict

from __future__ import annotations
import contextlib
import fnmatch
import glob
import logging
import os
import pathlib
import shutil
import sqlite3
import unittest
import warnings

from urllib.parse import urlparse

from netaddr import valid_ipv6

from .utils import getPortFromPath
from .utils import ManagedResource

from ..slap.standalone import StandaloneSlapOS
from ..slap.standalone import SlapOSNodeCommandError
from ..slap.standalone import PathTooDeepError

from ..util import mkdir_p
from ..slap import ComputerPartition
from .check_software import checkSoftware

from ..proxy.db_version import DB_VERSION

from typing import (
  Callable,
  ClassVar,
  Dict,
  Iterable,
  Iterator,
  Mapping,
  Sequence,
  Tuple,
  Type,
  TypeVar,
)

ManagedResourceType = TypeVar("ManagedResourceType", bound=ManagedResource)

IPV4_ADDRESS_DEFAULT: str = os.environ["SLAPOS_TEST_IPV4"]
IPV6_ADDRESS_DEFAULT: str = os.environ["SLAPOS_TEST_IPV6"]
DEBUG_DEFAULT: bool = bool(
  int(os.environ.get("SLAPOS_TEST_DEBUG", 0)),
)
VERBOSE_DEFAULT: bool = bool(
  int(os.environ.get("SLAPOS_TEST_VERBOSE", 0)),
)
SKIP_SOFTWARE_CHECK_DEFAULT: bool = bool(
  int(os.environ.get("SLAPOS_TEST_SKIP_SOFTWARE_CHECK", 0))
)
SKIP_SOFTWARE_REBUILD_DEFAULT: bool = bool(
  int(os.environ.get("SLAPOS_TEST_SKIP_SOFTWARE_REBUILD", 0))
)
SHARED_PART_LIST_DEFAULT: Sequence[str] = [
  os.path.expanduser(p)
  for p in os.environ.get(
    "SLAPOS_TEST_SHARED_PART_LIST",
    "",
  ).split(os.pathsep)
  if p
]
SNAPSHOT_DIRECTORY_DEFAULT: str | None = os.environ.get(
  "SLAPOS_TEST_LOG_DIRECTORY",
)


def makeModuleSetUpAndTestCaseClass(
  software_url: str | os.PathLike[str],
  *,
  base_directory: str | None = None,
  ipv4_address: str = IPV4_ADDRESS_DEFAULT,
  ipv6_address: str = IPV6_ADDRESS_DEFAULT,
  debug: bool = DEBUG_DEFAULT,
  verbose: bool = VERBOSE_DEFAULT,
  skip_software_check: bool = SKIP_SOFTWARE_CHECK_DEFAULT,
  skip_software_rebuild: bool = SKIP_SOFTWARE_REBUILD_DEFAULT,
  shared_part_list: Iterable[str] = SHARED_PART_LIST_DEFAULT,
  snapshot_directory: str | None = SNAPSHOT_DIRECTORY_DEFAULT,
  software_id: str | None = None,
) -> Tuple[Callable[[], None], Type[SlapOSInstanceTestCase]]:
  """
  Create a setup module function and a testcase for testing `software_url`.

  Note:
    SlapOS itself and some services running in SlapOS uses unix sockets and
    (sometimes very) deep paths, which do not play very well together.
    To workaround this, users can set ``SLAPOS_TEST_WORKING_DIR`` environment
    variable to the path of a short enough directory and local slapos will
    use this directory.
    The partitions references will be named after the unittest class name,
    which can also lead to long paths. For this, unit test classes can define
    a ``__partition_reference__`` attribute which will be used as partition
    reference. If the class names are long, the trick is then to use a shorter
    ``__partition_reference__``.
    See https://lab.nexedi.com/kirr/slapns for a solution to this problem.

  Args:
    software_url: The URL or path of the software to test.
    base_directory: The base directory used for SlapOS.
      By default, it will use the value in the environment variable
      ``SLAPOS_TEST_WORKING_DIR``.
      If that is not defined, it will default to ``.slapos`` in the current
      directory.
    ipv4_address: IPv4 address used for the instance. By default it will use
      the one defined in the environment variable ``SLAPOS_TEST_IPV4``.
    ipv6_address: IPv6 address used for the instance. By default it will use
      the one defined in the environment variable ``SLAPOS_TEST_IPV6``.
    debug: Enable debugging mode, which will drop in a debugger session when
      errors occur.
      By default it will be controlled by the value of the environment variable
      ``SLAPOS_TEST_DEBUG`` if it is defined. Otherwise it will be disabled.
    verbose: ``True`` to enable verbose logging, so that the test framework
      logs information describing the actions taken (sets logging level to
      ``DEBUG``).
      By default it will be controlled by the value of the environment variable
      ``SLAPOS_TEST_VERBOSE`` if it is defined. Otherwise it will be disabled.
    skip_software_check: Skips costly software checks.
      By default it will be controlled by the value of the environment variable
      ``SLAPOS_TEST_SKIP_SOFTWARE_CHECK`` if it is defined. Otherwise it will
      be disabled.
    skip_software_rebuild: Skips costly software builds.
      By default it will be controlled by the value of the environment variable
      ``SLAPOS_TEST_SKIP_SOFTWARE_REBUILD`` if it is defined. Otherwise it will
      be disabled.
    shared_part_list: Additional paths to search for existing shared parts.
      This test class will use its own directory for shared parts and also
      the paths defined in this argument.
      By default it is controlled by the ``SLAPOS_TEST_SHARED_PART_LIST``
      environment variable if defined, which should contain the paths in a
      string separated by colons (':').
    snapshot_directory: Directory to save snapshot files (for further
      inspection) and logs.
      If it is ``None`` or the empty string, logs will be stored in
      ``base_directory``, and no snapshots will be stored.
      By default it will use the value of the environment variable
      ``SLAPOS_TEST_LOG_DIRECTORY`` if it is defined, and ``None`` otherwise.
    software_id: A short name for the software, to be used in logs and to
      name the snapshots.
      By default it is computed automatically from the software URL, but can
      also be passed explicitly, to use a different name for different kind of
      tests, like for example upgrade tests.

  Returns:
    A tuple of two arguments:
      - A function to install the software, to be used as `unittest`'s
        `setUpModule`.
      - A base class for test cases.

  """
  software_url = os.fspath(software_url)

  if base_directory is None:
    base_directory = os.path.realpath(
      os.environ.get(
        "SLAPOS_TEST_WORKING_DIR",
        os.path.join(
          os.getcwd(),
          ".slapos",
        ),
      )
    )

  if not software_id:
    software_id = urlparse(software_url).path.split("/")[-2]

  logging.basicConfig(
    level=logging.DEBUG,
    format=f"%(asctime)s - {software_id} - %(name)s - %(levelname)s - %(message)s",
    filename=os.path.join(
      snapshot_directory or base_directory,
      "testcase.log",
    ),
  )
  logger = logging.getLogger()
  console_handler = logging.StreamHandler()
  console_handler.setLevel(
    logging.DEBUG if verbose else logging.WARNING,
  )
  logger.addHandler(console_handler)

  if debug:
    unittest.installHandler()

  # TODO: fail if already running ?
  try:
    slap = StandaloneSlapOS(
      base_directory=base_directory,
      server_ip=ipv4_address,
      server_port=getPortFromPath(base_directory),
      shared_part_list=shared_part_list,
    )
  except PathTooDeepError:
    raise RuntimeError(
      f"base directory ( {base_directory} ) is too deep, try setting "
      f"SLAPOS_TEST_WORKING_DIR to a shallow enough directory",
    )

  cls = type(
    f"SlapOSInstanceTestCase for {software_url}",
    (SlapOSInstanceTestCase,),
    {
      "slap": slap,
      "getSoftwareURL": classmethod(lambda _cls: software_url),
      "software_id": software_id,
      "_debug": debug,
      "_skip_software_check": skip_software_check,
      "_skip_software_rebuild": skip_software_rebuild,
      "_ipv4_address": ipv4_address,
      "_ipv6_address": ipv6_address,
      "_base_directory": base_directory,
      "_test_file_snapshot_directory": snapshot_directory,
    },
  )

  class SlapOSInstanceTestCase_(
    cls,
    SlapOSInstanceTestCase,
  ):
    # useless intermediate class so that editors provide completion anyway.
    pass

  def setUpModule() -> None:
    installSoftwareUrlList(cls, [software_url], debug=debug)

  return setUpModule, SlapOSInstanceTestCase_


def installSoftwareUrlList(
  cls: Type[SlapOSInstanceTestCase],
  software_url_list: Sequence[str],
  max_retry: int = 10,
  debug: bool = False,
) -> None:
  """
  Install softwares on the current testing slapos, for use in `setUpModule`.

  This also check softwares with `checkSoftware`.

  Args:
    cls: The test case class used for the installation.
    software_url_list: List of URLs or paths to install.
    max_retry: Number of times that the installation will be retried if there
      is an error.
    debug: If set to ``True`` the software will not be automatically removed
      if there is an error during the installation process, in order to
      facilitate inspection during debug.

  """

  def _storeSoftwareSnapshot(name: str) -> None:
    for path in (
      glob.glob(
        os.path.join(
          cls._base_directory,  # pyright: ignore[reportPrivateUsage]
          "var/log/*",
        )
      )
      + glob.glob(
        os.path.join(
          cls.slap.software_directory,
          "*/*.cfg",
        )
      )
      + glob.glob(
        os.path.join(
          cls.slap.software_directory,
          "*/.installed.cfg",
        )
      )
      + glob.glob(
        os.path.join(
          cls.slap.shared_directory,
          "*/*/.slapos.recipe.cmmi.signature",
        )
      )
    ):
      cls._copySnapshot(path, name)  # pyright: ignore[reportPrivateUsage]

  try:
    cls.logger.debug("Starting SlapOS")
    cls.slap.start()
    for software_url in software_url_list:
      cls.logger.debug("Supplying %s", software_url)
      cls.slap.supply(software_url)
    cls.logger.debug("Waiting for slapos node software to build")
    cls.slap.waitForSoftware(
      max_retry=max_retry,
      debug=debug,
      install_all=not cls._skip_software_rebuild,  # pyright: ignore[reportPrivateUsage]
    )
    _storeSoftwareSnapshot("setupModule")
    if not cls._skip_software_check:  # pyright: ignore[reportPrivateUsage]
      for software_url in software_url_list:
        cls.logger.debug("Checking software %s", software_url)
        checkSoftware(cls.slap, software_url)
        cls.logger.debug("Done checking software %s", software_url)
    else:
      cls.logger.debug("Software checks skipped")

  except BaseException:
    _storeSoftwareSnapshot("setupModule failed installing software")
    if not debug:
      cls.logger.exception("Error building software, removing")
      try:
        for software_url in software_url_list:
          cls.logger.debug("Removing %s", software_url)
          cls.slap.supply(software_url, state="destroyed")
        cls.logger.debug("Waiting for slapos node software to remove")
        cls.slap.waitForSoftware(max_retry=max_retry, debug=debug)
      except BaseException:
        cls.logger.exception("Error removing software")
        _storeSoftwareSnapshot("setupModule removing software")
    cls._cleanup("setupModule")  # pyright: ignore[reportPrivateUsage]
    raise


class SlapOSInstanceTestCase(unittest.TestCase):
  """
  Install one slapos instance.

  This test case install software(s) and request one instance
  during `setUpClass` and destroy that instance during `tearDownClass`.

  Software Release URL, Instance Software Type and Instance Parameters
  can be defined on the class.

  All tests from the test class will run with the same instance.

  Note:
    This class is not supposed to be imported directly, but needs to be setup
    by calling makeModuleSetUpAndTestCaseClass.

  Attributes:
    computer_partition: The computer partition instance.
    computer_partition_root_path: The path of the instance root directory.
    computer_partition_ipv6_address: The IPv6 of the instance.
    instance_max_retry: Maximum retries for ``slapos node instance``.
    report_max_retry: Maximum retries for ``slapos node report``.
    partition_count: Number of partitions needed for this instance.
    default_partition_reference: Reference of the default requested partition.
    request_instance: Whether an instance needs to be requested for this test
        case.
    software_id: A short name of that software URL. E.g. helloworld instead of
        https://lab.nexedi.com/nexedi/slapos/raw/software/helloworld/software.cfg .
    logger: A logger for messages of the testing framework.
    slap: Standalone SlapOS instance.
  """

  instance_max_retry: ClassVar[int] = 20
  report_max_retry: ClassVar[int] = 20
  partition_count: ClassVar[int] = 10
  default_partition_reference: ClassVar[str] = "testing partition 0"
  request_instance: ClassVar[bool] = True
  software_id: ClassVar[str] = ""

  logger: ClassVar[logging.Logger] = logging.getLogger(__name__)

  # Dynamic members
  slap: ClassVar[StandaloneSlapOS]
  computer_partition: ClassVar[ComputerPartition]
  computer_partition_root_path: ClassVar[pathlib.Path]
  computer_partition_ipv6_address: ClassVar[str]

  # Private settings

  # Partition reference: use when default length is too long.
  __partition_reference__: ClassVar[str]

  # True to enable debugging utilities.
  _debug: ClassVar[bool] = False

  # True to skip software checks.
  _skip_software_check: ClassVar[bool] = False

  # True to skip software rebuild.
  _skip_software_rebuild: ClassVar[bool] = False

  # The IPv4 address used by ``slapos node format``.
  _ipv4_address: ClassVar[str] = ""

  # The IPv6 address used by ``slapos node format``.
  _ipv6_address: ClassVar[str] = ""

  # Used resources.
  _resources: ClassVar[Dict[str, ManagedResource]] = {}

  # Instance parameters
  _instance_parameter_dict: ClassVar[Mapping[str, object]]

  # Base directory for standalone SlapOS.
  _base_directory: ClassVar[str] = ""

  # Directory to save snapshot files for inspections.
  _test_file_snapshot_directory: ClassVar[str | None] = ""

  # Patterns of files to save for inspection, relative to instance directory.
  _save_instance_file_pattern_list: ClassVar[Sequence[str]] = (
    "*/bin/*",
    "*/etc/*",
    "*/var/log/*",
    "*/srv/monitor/*",
    "*/srv/backup/logrotate/*",
    "*/.*log",
    "*/.*cfg",
    "*/*cfg",
    "etc/",
  )

  @classmethod
  def getManagedResource(
    cls,
    resource_name: str,
    resource_class: Type[ManagedResourceType],
  ) -> ManagedResourceType:
    """
    Get the managed resource for this name.

    If resource was not created yet, it is created and `open`. The
    resource will automatically be `close` at the end of the test
    class.

    Args:
      resource_name: The name of the resource.
      resource_class: The desired class of the resource. If the resource
        exists, but is not an instance of this class, an exception will be
        raised. Otherwise, if the resource does not exist, this class will be
        used to construct a new resource with that name.

    Returns:
      A resource with name ``resource_name`` and class ``resource_class``.

    """
    try:
      existing_resource = cls._resources[resource_name]
    except KeyError:
      resource = resource_class(cls, resource_name)
      cls._resources[resource_name] = resource
      resource.open()
      return resource
    else:
      if not isinstance(existing_resource, resource_class):
        raise ValueError(
          f"Resource {resource_name} is of unexpected "
          f"class {existing_resource}",
        )
      return existing_resource

  # Methods to be defined by subclasses.
  @classmethod
  def getSoftwareURL(cls) -> str:
    """
    Return URL of software release to request instance.

    This method will be defined when initialising the class
    with makeModuleSetUpAndTestCaseClass.

    Returns:
      URL of the software release to request.

    """
    raise NotImplementedError()

  @classmethod
  def getInstanceParameterDict(cls) -> Mapping[str, object]:
    """
    Return instance parameters.

    To be defined by subclasses if they need to request instance
    with specific parameters.

    Returns:
      A mapping with the parameters to be set in the instance.

    """
    return {}

  @classmethod
  def getInstanceSoftwareType(cls) -> str | None:
    """
    Return software type for instance, default None.

    To be defined by subclasses if they need to request instance with specific
    software type.

    Returns:
      Name of the software type, or `None` to use the default software type.

    """
    return None

  # Unittest methods
  @classmethod
  def waitForInstance(cls) -> None:
    """
    Wait for the instance to be ready.

    This method does retry several times until either the instance is ready or
    `cls.instance_max_retry` unsuccessful retries have been done.

    """
    # waitForInstance does not tolerate any error but with instances,
    # promises sometimes fail on first run, because services did not
    # have time to start.
    # To make debug usable, we tolerate instance_max_retry-1 errors and
    # only debug the last.
    if cls._debug and cls.instance_max_retry:
      try:
        cls.slap.waitForInstance(max_retry=cls.instance_max_retry - 1)
      except SlapOSNodeCommandError:
        cls.slap.waitForInstance(debug=True)
    else:
      cls.slap.waitForInstance(
        max_retry=cls.instance_max_retry,
        debug=cls._debug,
      )

  @classmethod
  def formatPartitions(cls) -> None:
    """Format the instance partitions."""
    cls.logger.debug(
      "Formatting to remove old partitions XXX should not be needed because we delete ..."
    )
    cls.slap.format(0, cls._ipv4_address, cls._ipv6_address)
    cls.logger.debug("Formatting with %s partitions", cls.partition_count)
    cls.slap.format(
      cls.partition_count,
      cls._ipv4_address,
      cls._ipv6_address,
      getattr(cls, "__partition_reference__", f"{cls.__name__}-"),
    )

  @classmethod
  def _setUpClass(cls) -> None:
    cls.slap.start()

    # (re)format partitions
    cls.formatPartitions()

    # request
    if cls.request_instance:
      cls.requestDefaultInstance()

      # slapos node instance
      cls.logger.debug("Waiting for instance")
      cls.waitForInstance()

      # expose some class attributes so that tests can use them:
      # the main ComputerPartition instance, to use getInstanceParameterDict
      cls.computer_partition = cls.requestDefaultInstance()

      # the path of the instance on the filesystem, for low level inspection
      cls.computer_partition_root_path = pathlib.Path(
        cls.slap._instance_root,  # pyright: ignore[reportPrivateUsage]
      ) / cls.computer_partition.getId()

      # the ipv6 of the instance
      cls.computer_partition_ipv6_address = cls.getPartitionIPv6(
        cls.computer_partition.getId(),
      )

  @classmethod
  @contextlib.contextmanager
  def _snapshotManager(cls, snapshot_name: str) -> Iterator[None]:
    try:
      yield
    except BaseException:
      cls._storeSystemSnapshot(snapshot_name)
      cls._cleanup(snapshot_name)
      raise
    else:
      cls._storeSystemSnapshot(snapshot_name)

  @classmethod
  def setUpClass(cls):
    """Request an instance."""
    cls.logger.debug("Starting setUpClass %s", cls)
    cls._instance_parameter_dict = cls.getInstanceParameterDict()
    snapshot_name = "{}.{}.setUpClass".format(cls.__module__, cls.__name__)

    with cls._snapshotManager(snapshot_name):
      try:
        cls._setUpClass()
      except BaseException:
        cls.logger.exception("Error during setUpClass")
        cls.setUp = lambda self: self.fail("Setup Class failed.")
        raise
    cls.logger.debug("setUpClass done")

  @classmethod
  def tearDownClass(cls):
    """Tear down class, stop the processes and destroy instance."""
    cls._cleanup(f"{cls.__module__}.{cls.__name__}.tearDownClass")
    if not cls._debug:
      cls.logger.debug(
        "cleaning up slapos log files in %s",
        cls.slap._log_directory,  # pyright: ignore[reportPrivateUsage]
      )
      for log_file in glob.glob(
        os.path.join(
          cls.slap._log_directory,  # pyright: ignore[reportPrivateUsage]
          "*",
        )
      ):
        os.unlink(log_file)

  @classmethod
  def _storePartitionSnapshot(cls, name: str) -> None:
    """
    Store snapshot of partitions.

    This uses the definition from class attribute
    `_save_instance_file_pattern_list`.

    Args:
      name: Name of the snapshot.

    """
    # copy config and log files from partitions
    for dirpath, dirnames, filenames in os.walk(cls.slap.instance_directory):
      for dirname in list(dirnames):
        dirabspath = os.path.join(dirpath, dirname)
        if any(
          fnmatch.fnmatch(
            dirabspath,
            pattern,
          )
          for pattern in cls._save_instance_file_pattern_list
        ):
          cls._copySnapshot(dirabspath, name)
          # don't recurse, since _copySnapshot is already recursive
          dirnames.remove(dirname)
      for filename in filenames:
        fileabspath = os.path.join(dirpath, filename)
        if any(
          fnmatch.fnmatch(
            fileabspath,
            pattern,
          )
          for pattern in cls._save_instance_file_pattern_list
        ):
          cls._copySnapshot(fileabspath, name)

  @classmethod
  def _storeSystemSnapshot(cls, name: str) -> None:
    """
    Store a snapshot of standalone slapos and partitions.

    Does not include software log, because this is stored at the end of
    software installation and software log is large.

    Args:
      name: Name of the snapshot.

    """
    # copy log files from standalone
    for standalone_log in glob.glob(
      os.path.join(
        cls._base_directory,
        "var/log/*",
      )
    ):
      if not standalone_log.startswith("slapos-node-software.log"):
        cls._copySnapshot(standalone_log, name)
    # store slapproxy database
    cls._copySnapshot(
      cls.slap._proxy_database,  # pyright: ignore[reportPrivateUsage]
      name,
    )
    cls._storePartitionSnapshot(name)

  def tearDown(self):
    self._storePartitionSnapshot(self.id())

  @classmethod
  def _copySnapshot(cls, source_file_name: str, name: str) -> None:
    """
    Save a file, symbolic link or directory for later inspection.

    The path are made relative to slapos root directory and
    we keep the same directory structure.

    Args:
      source_file_name: The name of the file or directory to copy.
      name: Name of the snapshot.

    """
    if not cls._test_file_snapshot_directory:
      warnings.warn("No snapshot directory configured, skipping snapshot")
      warnings.warn(
        "Snapshot directory can be configured with SLAPOS_TEST_LOG_DIRECTORY environment"
      )
      return
    # we cannot use os.path.commonpath on python2, so implement something similar
    common_path = os.path.commonprefix((source_file_name, cls._base_directory))
    if not os.path.isdir(common_path):
      common_path = os.path.dirname(common_path)

    relative_path = source_file_name[len(common_path) :]
    if relative_path[0] == os.sep:
      relative_path = relative_path[1:]
    destination = os.path.join(
      cls._test_file_snapshot_directory,
      cls.software_id,
      name,
      relative_path,
    )
    destination_dirname = os.path.dirname(destination)
    mkdir_p(destination_dirname)
    if os.path.islink(source_file_name) and not os.path.exists(
      source_file_name
    ):
      cls.logger.debug(
        "copy broken symlink %s as %s",
        source_file_name,
        destination,
      )
      with open(destination, "w") as f:
        f.write(f"broken symink to {os.readlink(source_file_name)}\n")
    elif os.path.isfile(source_file_name):
      shutil.copy(source_file_name, destination)
    elif os.path.isdir(source_file_name):
      # we copy symlinks as symlinks, so that this does not fail when
      # we copy a directory containing broken symlinks.
      shutil.copytree(source_file_name, destination, symlinks=True)

  # implementation methods
  @classmethod
  def _cleanup(cls, snapshot_name: str) -> None:
    """
    Destroy all instances and stop subsystem.

    Catches and log all exceptions and take snapshot named `snapshot_name` + the failing step.

    Args:
      snapshot_name: Name of the snapshot that will be taken in case of exception.

    """
    for resource_name in list(cls._resources):
      cls.logger.debug("closing resource %s", resource_name)
      try:
        cls._resources.pop(resource_name).close()
      except:
        cls.logger.exception("Error closing resource %s", resource_name)
    try:
      if cls.request_instance and hasattr(cls, "_instance_parameter_dict"):
        cls.requestDefaultInstance(state="destroyed")
    except:
      cls.logger.exception("Error during request destruction")
      cls._storeSystemSnapshot(f"{snapshot_name}._cleanup request destroy")
    try:
      # To make debug usable, we tolerate report_max_retry-1 errors and
      # only debug the last.
      for _ in range(3):
        if cls._debug and cls.report_max_retry:
          try:
            cls.slap.waitForReport(max_retry=cls.report_max_retry - 1)
          except SlapOSNodeCommandError:
            cls.slap.waitForReport(debug=True)
        else:
          cls.slap.waitForReport(
            max_retry=cls.report_max_retry, debug=cls._debug
          )
    except:
      cls.logger.exception("Error during actual destruction")
      cls._storeSystemSnapshot(f"{snapshot_name}._cleanup waitForReport")
    leaked_partitions = [
      cp
      for cp in cls.slap.computer.getComputerPartitionList()
      if cp.getState() != "destroyed"
    ]
    if leaked_partitions:
      cls.logger.critical(
        "The following partitions were not cleaned up: %s",
        [cp.getId() for cp in leaked_partitions],
      )
      cls._storeSystemSnapshot(
        "{}._cleanup leaked_partitions".format(snapshot_name)
      )
      for cp in leaked_partitions:
        try:
          # XXX is this really the reference ?
          partition_reference = cp.getInstanceParameterDict()["instance_title"]
          assert isinstance(partition_reference, str)
          cls.slap.request(
            software_release=cp.getSoftwareRelease().getURI(),
            # software_type=cp.getType(), # TODO
            partition_reference=partition_reference,
            state="destroyed",
          )
        except:
          cls.logger.exception(
            "Error during request destruction of leaked partition",
          )
          cls._storeSystemSnapshot(
            f"{snapshot_name}._cleanup leaked_partitions request destruction",
          )
      try:
        # To make debug usable, we tolerate report_max_retry-1 errors and
        # only debug the last.
        for _ in range(3):
          if cls._debug and cls.report_max_retry:
            try:
              cls.slap.waitForReport(max_retry=cls.report_max_retry - 1)
            except SlapOSNodeCommandError:
              cls.slap.waitForReport(debug=True)
          else:
            cls.slap.waitForReport(
              max_retry=cls.report_max_retry,
              debug=cls._debug,
            )
      except:
        cls.logger.exception(
          "Error during leaked partitions actual destruction",
        )
        cls._storeSystemSnapshot(
          f"{snapshot_name}._cleanup leaked_partitions waitForReport",
        )
    try:
      cls.slap.stop()
    except:
      cls.logger.exception("Error during stop")
      cls._storeSystemSnapshot(f"{snapshot_name}._cleanup stop")
    leaked_supervisor_configs = glob.glob(
      os.path.join(
        cls.slap.instance_directory,
        "etc/supervisord.conf.d/*.conf",
      )
    )
    if leaked_supervisor_configs:
      for config in leaked_supervisor_configs:
        os.unlink(config)
      raise AssertionError(
        f"Test leaked supervisor configurations: {leaked_supervisor_configs}",
      )

  @classmethod
  def requestDefaultInstance(
    cls,
    state: str = "started",  # TODO: Change to enum/Literal when all code is Python 3.
  ) -> ComputerPartition:
    software_url = cls.getSoftwareURL()
    software_type = cls.getInstanceSoftwareType()
    cls.logger.debug(
      'requesting "%s" software:%s type:%r state:%s parameters:%s',
      cls.default_partition_reference,
      software_url,
      software_type,
      state,
      cls._instance_parameter_dict,
    )
    return cls.slap.request(
      software_release=software_url,
      software_type=software_type,
      partition_reference=cls.default_partition_reference,
      partition_parameter_kw=cls._instance_parameter_dict,
      state=state,
    )

  @classmethod
  def getPartitionId(cls, instance_name: str) -> str:
    """
    Get the id of the partition.

    Args:
      instance_name: Name of the instance.

    Returns:
      Id of the partition.

    """
    query = (
      f"SELECT reference FROM partition{DB_VERSION} "
      f"WHERE partition_reference=?"
    )
    with sqlite3.connect(
      os.path.join(
        cls._base_directory,
        "var/proxy.db",
      )
    ) as db:
      return db.execute(query, (instance_name,)).fetchall()[0][0]

  @classmethod
  def getPartitionIPv6(cls, partition_id: str) -> str:
    """
    Get the IP address of the partition.

    Args:
      partition_id: Id of the partition.

    Returns:
      An IPv6 address in presentation (string) format.

    """
    query = (
      f"SELECT address FROM partition_network{DB_VERSION} "
      f"WHERE partition_reference=?"
    )
    with sqlite3.connect(
      os.path.join(
        cls._base_directory,
        "var/proxy.db",
      )
    ) as db:
      rows = db.execute(query, (partition_id,)).fetchall()
    # do not assume the partition's IPv6 address is the second one,
    # instead find the first address that is IPv6
    for (address,) in rows:
      if valid_ipv6(address):
        return address

    raise ValueError("Missing IPv6 address")



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2010-2014 Vifib SARL and Contributors.
# All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation; either version 2.1
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################

from six.moves import configparser
import os

from slapos.cli.command import Command


class ConfigError(Exception):
    pass


class ConfigCommand(Command):
    """
    Base class for commands that require a configuration file
    """

    default_config_var = 'SLAPOS_CONFIGURATION'

    # use this if default_config_var does not exist
    default_config_path = '/etc/opt/slapos/slapos.cfg'

    def get_parser(self, prog_name):
        ap = super(ConfigCommand, self).get_parser(prog_name)
        ap.add_argument('--cfg',
                        help='SlapOS configuration file'
                             ' (default: $%s or %s)' %
                             (self.default_config_var, self.default_config_path))
        return ap

    def config_path(self, args):
        if args.cfg:
            cfg_path = args.cfg
        else:
            cfg_path = os.environ.get(self.default_config_var, self.default_config_path)
        return os.path.expanduser(cfg_path)

    def fetch_config(self, args):
        """
        Returns a configuration object if file exists/readable/valid,
        will raise an error otherwise. The exception may come from the
        configparser itself if the configuration content is very broken,
        and will clearly show what is wrong with the file.
        """

        cfg_path = self.config_path(args)

        self.app.log.debug('Loading config: %s', cfg_path)

        if not os.path.exists(cfg_path):
            raise ConfigError('Configuration file does not exist: %s' % cfg_path)

        configp = configparser.ConfigParser()
        if configp.read(cfg_path) != [cfg_path]:
            # bad permission, etc.
            raise ConfigError('Cannot parse configuration file: %s' % cfg_path)

        return configp


class ClientConfigCommand(ConfigCommand):
    """
    Base class for client commands, that use the client configuration file
    """

    default_config_var = 'SLAPOS_CLIENT_CONFIGURATION'
    default_config_path = '~/.slapos/slapos-client.cfg'
    command_group = 'client'
